#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it 
# and/or modify it under the terms of the GNU General Public 
# License as published by the Free Software Foundation; either 
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be 
# useful, but WITHOUT ANY WARRANTY; without even the implied 
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public 
# License along with program; see the file COPYING. If not, 
# write to the Free Software Foundation, Inc., 59 Temple Place 
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000, 2001 Free Software Foundation
#
# FILE:
# oracle/DBdriver.py
#
# DESCRIPTION:
# Driver to provide access to data via Digital Creations Oracle/Python Driver
# Requires DCOracle2 2.x (http://www.zope.org/Products/DCOracle2/)
#
# NOTES:
#
#   Supported attributes (via connections.conf or <database> tag)
#
#     service=    This is the Oracle TNS name for your connection  (required)
#

import string
import sys
from gnue.common import GDebug, GDataObjects, GConnections
from gnue.common.dbdrivers._dbsig.DBdriver \
   import DBSIG_RecordSet, DBSIG_ResultSet, DBSIG_DataObject, \
          DBSIG_DataObject_SQL, DBSIG_DataObject_Object

try:
  import DCOracle2 as SIG2api
except ImportError, message:
  raise GConnections.AdapterNotInstalled, \
         "Driver not installed: DCOracle2 [%s]" % message



class Oracle_RecordSet(DBSIG_RecordSet):
  pass


class Oracle_ResultSet(DBSIG_ResultSet):
  def __init__(self, dataObject, cursor=None, defaultValues={}, masterRecordSet=None):
    DBSIG_ResultSet.__init__(self, dataObject, \
            cursor, defaultValues, masterRecordSet)
    self._recordSetClass = Oracle_RecordSet
    self._uniqueIdField = "__GNUeF__uniqueKey_%s" % (self._dataObject.table)
    self._uniqueIdFormat = "ROWID='%s'"



class Oracle_DataObject(DBSIG_DataObject):
  def __init__(self):
    DBSIG_DataObject.__init__(self)
    self._DatabaseError = SIG2api.DatabaseError
    self._resultSetClass = Oracle_ResultSet


  def connect(self, connectData={}):
    GDebug.printMesg(1,"Oracle database driver initializing")
    try:
      self._dataConnection = SIG2api.connect("%s/%s@%s" \
                % (connectData['_username'], \
                   connectData['_password'], \
                   connectData['service']))
    except self._DatabaseError, value:
      raise GDataObjects.LoginError, value

    self._postConnect()


  def _postConnect(self):
    self.triggerExtensions = TriggerExtensions(self._dataConnection)


  #
  # Schema (metadata) functions
  #

  # Return a list of the types of Schema objects this driver provides
  def getSchemaTypes(self):
    return [ ('user_table',   'User Table',1),
             ('user_view',    'User View',1), 
             ('user_synonym', 'User Synonym',1),
             ('all_table',    'System Table',1),
             ('all_view',     'System View',1),
             ('all_synonym',  'System Synonym',1) ]

  # Return a list of Schema objects
  def getSchemaList(self, type=None):

    where_user = ""
    if type == None: 
      where_type = ['TABLE', 'VIEW', 'SYNONYM']
    else: 
      scope, type = string.split(type,'_')
      where_type = [string.upper(type)]
      if scope == 'user': 
        where_user = " AND OWNER = USER"
    

    statement = \
      "select owner||'.'||table_name||'.'||table_type full_name, " + \
        "decode(owner,user,null,owner||'.')||table_name table_name, " + \
        "decode(owner,user,'user_','all_')||lower(table_type) table_type " + \
        "from all_catalog where table_type in ('%s') %s " \
              % (string.join(where_type,"','"), where_user) + \
          "order by table_name "

    GDebug.printMesg(5,statement)

    cursor = self._dataConnection.cursor()
    cursor.execute(statement)

    list = []
    for rs in cursor.fetchall():
      list.append(GDataObjects.Schema(attrs={'id':string.lower(rs[0]), 'name':rs[1],
                         'type':rs[2]},
                         getChildSchema=self.__getFieldSchema))

    cursor.close()
    return list


  # Find a schema object with specified name
  def getSchemaByName(self, name, type=None):
     return None


  # Get fields for a table
  def __getFieldSchema(self, parent):

    owner, name, type = string.split(parent.id,'.')

    cursor = self._dataConnection.cursor()

    if type == 'SYNONYM': 
      statement = "select table_owner, table_name, " + \
                  "decode(db_link,null,null,'@'||db_link) name " + \
                  "from all_synonyms " + \
                  "where owner = '%s' and synonym_name='%s'" % (owner, name)

      GDebug.printMesg(5,statement)

      cursor.execute(statement)
      rs = cursor.fetchone()
      owner, name, link = rs
    else: 
      link = ""

    statement = \
       "select owner||'.'||table_name||'.'||column_name||'.%s', " % (link) + \
       "column_name, data_type, nullable, data_length, data_scale " + \
       "from all_tab_columns%s " % (link) + \
       "where owner = '%s' and table_name = '%s' " % (owner, name) + \
       "order by column_id"

    GDebug.printMesg(5,statement)

    cursor.execute(statement)
  
    list = []
    for rs in cursor.fetchall():

      attrs={'id': rs[0], 'name': rs[1],
             'type':'field', 'nativetype': rs[2],
             'required': rs[3] == 'Y'}

      if rs[2] in ('NUMBER','LONG','LONG RAW'):
        attrs['precision'] = rs[5]
        attrs['datatype'] = 'number'
      elif rs[2] in ('DATE'):
        attrs['datatype'] = 'date'
      else:
        attrs['datatype'] = 'text'

      if rs[5] != 0:
        attrs['length'] = rs[4]

      list.append(GDataObjects.Schema(attrs=attrs))

    cursor.close()
    return tuple(list)


class Oracle_DataObject_Object(Oracle_DataObject, \
      DBSIG_DataObject_Object):

  def __init__(self):
    # Call DBSIG init first because Oracle_DataObject needs to overwrite
    # some of its values
    DBSIG_DataObject_Object.__init__(self)
    Oracle_DataObject.__init__(self)

  def _buildQuery(self, conditions={}):
    return DBSIG_DataObject_Object._buildQuery(self, conditions)


class Oracle_DataObject_SQL(Oracle_DataObject, \
      DBSIG_DataObject_SQL): 
  def __init__(self): 
    # Call DBSIG init first because Oracle_DataObject needs to overwrite 
    # some of its values
    DBSIG_DataObject_SQL.__init__(self) 
    Oracle_DataObject.__init__(self)

  def _buildQuery(self, conditions={}):
    return DBSIG_DataObject_SQL._buildQuery(self, conditions)


#
#  Extensions to Trigger Namespaces
#
class TriggerExtensions:

  def __init__(self, connection):
    self.__connection = connection

  # Return the current date, according to database
  def getTimeStamp(self):
    return self.__singleQuery("select sysdate from dual")

  # Return a sequence number from sequence 'name'
  def getSequence(self, name):
    return self.__singleQuery("select %s.nextval from dual" % name)

  # Run the SQL statement 'statement'
  def sql(self, statement):
    cursor = self.__connection.cursor()
    try:
      cursor.execute(statement)
      cursor.close()
    except:
      cursor.close()
      raise

  # Used internally
  def __singleQuery(self, statement):
    cursor = self.__connection.cursor()
    try:
      cursor.execute(statement)
      rv = cursor.fetchone()
      cursor.close()
    except mesg:
      GDebug.printMesg(1,"**** Unable to execute extension query")
      GDebug.printMesg(1,"**** %s" % mesg)
      cursor.close()
      return None

    try:
      return rv[0]
    except:
      return None




######################################
#
#  The following hashes describe
#  this driver's characteristings.
#
######################################

#
#  All datasouce "types" and corresponding DataObject class
#
supportedDataObjects = {
  'object': Oracle_DataObject_Object,
  'sql':    Oracle_DataObject_SQL
}



